import json
from collections import defaultdict
from logging import getLogger
from os import PathLike
from pathlib import Path
from typing import Dict, Optional

from interactions.client.context import _Context

from interactions import MISSING, Client, Locale

__all__ = ("Localization",)

log = getLogger("i18n")

# TODO:
#   Implement way to add/update json files


class Localization:
    def __init__(self, client: Client):
        self.client: Client = client
        self._locales: Dict[str, Dict[Locale, str]] = defaultdict(dict)

    @property
    def locales(self) -> Dict[str, Dict[Locale, str]]:
        """
        Returns dict of loaded localizations
        """
        return self._locales

    def load(self, path: PathLike):
        """
        Loads file/files with localization

        :param PathLike path: The path to file or to folder
        """
        _path = Path(path)

        if _path.is_file():
            locale_data = self.__load_file(_path)
            if not locale_data:
                return
            self.__load_localization(_path, locale_data)

        for file in _path.iterdir():
            locale_data = self.__load_file(file)
            if not locale_data:
                continue

            self.__load_localization(file, locale_data)

    def __load_file(self, file: Path) -> Optional[dict]:
        """
        Serializes file into dict
        """
        if file.suffix != ".json":
            log.debug(f"File {file} is not json format. Skipping...")
            return

        with open(file, "r", encoding="utf-8") as _file:
            try:
                return json.load(_file)
            except json.JSONDecodeError:
                log.exception(f"Couldn't load file {file}!")

    def __load_localization(self, file: Path, locale_data: dict):
        """
        Serializes dict to locale dict
        """
        locale = Locale(file.name.removesuffix(".json"))
        for key, value in locale_data.items():
            self._locales[key] |= {locale: value}

    def get(self, key: str) -> Optional[Dict[Locale, str]]:
        """
        Gets dict with localized value.

        :param str key: The key to get value.
        :return: Dict with Locale as key and string as value
        """
        return self.locales.get(key.upper())

    def get_translate(self, locale: Locale, key: str) -> Optional[str]:
        """
        Gets a translation from locale code and key.

        :param Locale locale: The locale code.
        :param str key: The key to get localization
        :return: localized string, if any.
        """

        _localizations: Dict[Locale, str] = self.get(key)
        return _localizations.get(locale) if _localizations else None

    def _get_name(self, key: str) -> Optional[Dict[Locale, str]]:
        """
        Gets dict with localized names to command/option name.

        :param str key: The key to get value.
        :return: Dict with Locale as key and string as value
        """
        _key: str = f"{key}_NAME"
        value = self.get(_key)
        return value if value is not None else MISSING

    def _get_description(self, key: str) -> Optional[Dict[Locale, str]]:
        """
        Gets dict with localized description to command/option name.

        :param str key: The key to get value.
        :return: Dict with Locale as key and string as value
        """
        _key: str = f"{key}_DESCRIPTION"
        value = self.get(_key)
        return value if value is not None else MISSING

    # TODO: Remove these methods after 4.4.0 release

    @staticmethod
    def get_locale(ctx: _Context) -> Optional[Locale]:
        """
        Hack to get locale from ctx.

        :param ctx:
        :return:
        """
        if hasattr(ctx, "locale"):
            return ctx.locale

        _locale = ctx._extras.get("locale")
        return Locale(_locale) if _locale else None

    @staticmethod
    def get_guild_locale(ctx: _Context) -> Optional[Locale]:
        """
        Hack to get guild locale from ctx.

        :param ctx:
        :return:
        """
        if hasattr(ctx, "guild_locale"):
            return ctx.guild_locale

        _guild_locale = ctx._extras.get("guild_locale")
        return Locale(_guild_locale) if _guild_locale else None



from typing import Awaitable, Callable, List, Optional, Union

from interactions.client.decor import command
from interactions.utils.attrs_utils import define, field

import interactions
from interactions import MISSING, Command
from interactions import option as _option

from .extension import Localization


def add_option_localizations(i18n: Localization, __option: interactions.Option):
    key = __option.name
    if locale_key := getattr(__option, "locale_key", None):
        key = locale_key

    _name = i18n._get_name(key)
    _description = i18n._get_description(key)

    if _name:
        __option.name_localizations = _name
        __option._json["name_localizations"] = _name  # TODO: Remove this after asdict pr
    if _description:
        __option.description_localizations = _description
        __option._json[
            "description_localizations"
        ] = _description  # TODO: Remove this after asdict pr


@property
def full_data(self) -> Union[dict, List[dict]]:
    i18n: Localization = self.client.i18n  # type: ignore

    self.name_localizations = i18n._get_name(self.name)
    self.description_localizations = i18n._get_description(self.name)

    for option in self.options:
        add_option_localizations(i18n, option)
        if option.options:
            for _option in option.options:  # noqa: F402
                add_option_localizations(i18n, _option)
                if _option.options:
                    for __option in _option.options:
                        add_option_localizations(i18n, __option)

    return command(
        type=self.type,
        name=self.name,
        description=self.description if self.type == 1 else MISSING,
        options=self.options if self.type == 1 else MISSING,
        scope=self.scope,
        name_localizations=self.name_localizations,
        description_localizations=self.description_localizations,
        default_member_permissions=self.default_member_permissions,
        dm_permission=self.dm_permission,
    )


setattr(Command, "full_data", full_data)


@define()
class Option(interactions.Option):
    locale_key: Optional[str] = field(default=None)


def option(
    description: str = "No description set",
    /,
    *,
    key: Optional[str] = None,
    **kwargs,
) -> Callable[[Callable[..., Awaitable]], Callable[..., Awaitable]]:
    def wrapper(coro: Callable[..., Awaitable]) -> Callable[..., Awaitable]:
        _coro = _option(description, **kwargs)(coro)

        # I'm getting AttributeError when trying to set attribute
        __option = Option(**_coro._options[0]._json)
        __option.locale_key = key
        _coro._options[0] = __option

        return _coro

    return wrapper


interactions.option = option
